module SignedDistanceField


using FileIO
using GeometryBasics
using LinearAlgebra
using StructArrays
using GLMakie
using ProgressMeter
using NearestNeighbors
using WriteVTK


include("meshutils.jl")

function compute_incident_angles(tripos)

    a = tripos[1]
    b = tripos[2]
    c = tripos[3]
    sab = dot(b - a, b - a)
    sbc = dot(b - c, b - c)
    sac = dot(c - a, c - a)
    α1 = acos((sab + sac - sbc) / (2 * sqrt(sab * sac)))
    α2 = acos((sbc + sab - sac) / (2 * sqrt(sbc * sab)))
    α3 = acos((sac + sbc - sab) / (2 * sqrt(sac * sbc)))

    return α1, α2, α3

end



function preprocess_mesh(mesh)

    positions = coordinates(mesh)
    trisconn = decompose(TriangleFace{Int}, mesh)
    edgepts, edges_in_tri = populate_edges(trisconn)


    points = StructArray(create_single_vertex(i, positions[i]) for i in eachindex(positions))
    edges = StructArray(create_single_edge(i, points, edgepts[i]) for i in eachindex(edgepts))
    triangles = StructArray(create_single_triangle(i, points, trisconn[i], edges, edges_in_tri[i]) for i in eachindex(trisconn))

    # Compute edge and point pseudo-normals
    # See Baerentzen and Aanaes (2002)
    for i in eachindex(triangles)
        n = triangles.normal[i]
        triangles.edges[i].normal += Vec3(n, n, n)
        α1, α2, α3 = compute_incident_angles(triangles.points[i].position)
        triangles.points[i].normal += Vec3(α1 * n, α2 * n, α3 * n)
    end

    return points, edges, triangles

end


"""
    squared_distance, sign = signed_distance(point, tripts, vertnorm, edgenorm, trinorm)

Closest distance between a point and a triangular face defined by:
- `tripts`: vertex coordinates
- `vertnorm`: point normals
- `edgenorm`: edge normals
- `trinorm`: face normal

Returns the squared distance and the sign.

Reference: Christer Ericson’s Real-time Collision Detection book.
"""
@fastmath function signed_distance(point, tripts, vertnorm, edgenorm, trinorm)

    a = tripts[1]
    b = tripts[2]
    c = tripts[3]

    na = vertnorm[1]
    nb = vertnorm[2]
    nc = vertnorm[3]

    nab = edgenorm[1]
    nbc = edgenorm[2]
    nac = edgenorm[3]

    n = trinorm

    p = point

    ab = b - a
    ac = c - a
    bc = c - b
    ap = p - a
    bp = p - b
    cp = p - c

    snom = dot(ap, ab)
    sdenom = dot(bp, -ab)

    tnom = dot(ap, ac)
    tdenom = dot(cp, -ac)

    # Closest point is vertex 1
    if snom <= 0 && tnom <= 0
        return dot(ap, ap), sign(dot(ap, na))
    end

    unom = dot(bp, bc)
    udenom = dot(cp, -bc)

    # Closest point is vertex 2
    if sdenom <= 0 && unom <= 0
        return dot(bp, bp), sign(dot(bp, nb))
    end

    # Closest point is vertex 3
    if tdenom <= 0 && udenom <= 0
        return dot(cp, cp), sign(dot(cp, nc))
    end

    vc = dot(n, cross(-ap, -bp))

    # Closest point is edge 12
    if vc <= 0 && snom >= 0 && sdenom >= 0
        v = p - (a + snom / (snom + sdenom) * ab)
        return dot(v, v), sign(dot(v, nab))
    end

    va = dot(n, cross(-bp, -cp))

    # Closest point is edge 23
    if va <= 0 && unom >= 0 && udenom >= 0
        v = p - (b + unom / (unom + udenom) * bc)
        return dot(v, v), sign(dot(v, nbc))
    end

    vb = dot(n, cross(-cp, -ap))

    # Closest point is edge 13
    if vb <= 0 && tnom >= 0 && tdenom >= 0
        v = p - (a + tnom / (tnom + tdenom) * ac)
        return dot(v, v), sign(dot(v, nac))
    end

    # Closest point is on the face
    denom2 = 1 / (va + vb + vc)
    v = vb * denom2
    w = vc * denom2
    d = p - (a + ab * v + ac * w)

    return dot(d, d), sign(dot(d, n))

end


"""
    d = minimum_signed_distance(point, triangles)

Compute the minimum distance between a point and a triangular mesh defined by a `StructArray` of triangles.
"""
function minimum_signed_distance(point, triangles)
    s = minimum(
        signed_distance(point, triangles.points[i].position, triangles.points[i].normal, triangles.edges[i].normal, triangles.normal[i]) 
        for i in eachindex(triangles)
    )
    return sqrt(s[1]) * s[2]
end


"""
    s = signed_distance_field(mesh, xaxis, yaxis, zaxis, acc)

Compute the signed distance field on `xaxis`, `yaxis` and `zaxis` grid points.
`acc` is the acceleration method (`Val(nothing)` or `Val(:KDTree)`).
"""
function signed_distance_field(mesh, xaxis, yaxis, zaxis, acc::Val{nothing})

    _, _, triangles = preprocess_mesh(mesh)

    s = Array{Float32,3}(undef, length(xaxis), length(yaxis), length(zaxis))

    p = Progress(length(zaxis), 1, "Computing signed distance field... ")
    @inbounds Threads.@threads for k in eachindex(zaxis)
        @inbounds for j in eachindex(yaxis)
            @inbounds for i in eachindex(xaxis)
                s[i, j, k] = minimum_signed_distance(Vec3(xaxis[i], yaxis[j], zaxis[k]), triangles)
            end
        end
        next!(p)
    end

    return s
end


#KDTree accelerated
function signed_distance_field(mesh, xaxis, yaxis, zaxis, acc::Val{:KDTree})

    _, _, triangles = preprocess_mesh(mesh)

    tricenters = [sum(p.position)/length(p.position) for p in triangles.points]

    candidates_tree = KDTree(tricenters; leafsize = 10)

    s = Array{Float32,3}(undef, length(xaxis), length(yaxis), length(zaxis))

    p = Progress(length(zaxis), 1, "Computing signed distance field... ")
    @inbounds Threads.@threads for k in eachindex(zaxis)
        @inbounds for j in eachindex(yaxis)
            @inbounds for i in eachindex(xaxis)
                point = Vec3(xaxis[i], yaxis[j], zaxis[k])
                candidates_idxs, _ = knn(candidates_tree, point, 10, false)
                stri = minimum(
                    signed_distance(point, triangles.points[i].position, triangles.points[i].normal, triangles.edges[i].normal, triangles.normal[i]) 
                    for i in candidates_idxs
                )
                s[i, j, k] = sqrt(stri[1]) * stri[2]
            end
        end
        next!(p)
    end

    return s
end




"""
    s, x, y, z = sdfgen(meshfile, dx; padding=1, acceleration=nothing)
	
Generate a signed distance field for a 3D triangular mesh at gridded points.

# Arguments
- `meshfile`: triangular mesh file (all files readable by the `MeshIO` package are accepted).
- `dx`: grid spacing.
- `padding` (optional): number of steps with which to pad the space.
- `acceleration` (optional): acceleration method based on space partition. Helps when the number of points is large. Can be `nothing` or `:KDTree`.
Note: KDTree acceleration assumes a rather homogeneous mesh size. The initial guesses for closest triangle
is based on the triangle barycenter position, and the refined search is performed on these 10 initial
guesses.

Returns the signed distance field `s` and the `x`, `y` and `z` grid axes corresponding to the evaluation points.

The algorithm is based on pseudo-normals, and assumes a watertight mesh.
It is described in: J. Andreas Baerentzen and Henrik Aanaes: Generating Signed Distance Fields From
Triangle Meshes, Technical Report, IMM. 2002.
"""
function sdfgen(meshfile, dx; padding = 1, acceleration=nothing)

    mesh = load(meshfile)

    min_box = Vec3f([minimum(v[i] for v in mesh.position) for i = 1:3] .- padding * dx)
    max_box = Vec3f([maximum(v[i] for v in mesh.position) for i = 1:3] .+ padding * dx)

    xaxis = min_box[1]:Float32(dx):max_box[1]
    yaxis = min_box[2]:Float32(dx):max_box[2]
    zaxis = min_box[3]:Float32(dx):max_box[3]

    @info "The bounding box of the mesh is: $(xaxis[1]):$(xaxis[end]), $(yaxis[1]):$(yaxis[end]), $(zaxis[1]):$(zaxis[end])"
    @info "Signed distance will be computed on a $(length(xaxis)) x $(length(yaxis)) x $(length(zaxis)) grid"

    s = signed_distance_field(mesh, xaxis, yaxis, zaxis, Val(acceleration))

    return s, xaxis, yaxis, zaxis

end



"""
    display_sdf(s, x, y, z)
	
Displays the signed distance field `s` on the grid defined by `x`, `y` and `z` axes.
"""
function display_sdf(s, x, y, z)

    fig = Figure()
    ax = Axis3(fig[1, 1], aspect = :data, viewmode = :fit, limits=(extrema(x)..., extrema(y)..., extrema(z)...))
    hidedecorations!(ax)
    hidespines!(ax)


    lsgrid = labelslidergrid!(fig, ["yz plane - x axis", "xz plane - y axis", "xy plane - z axis"], [x, y, z])
    fig[2, 1:2] = lsgrid.layout

    minall = minimum(s)
    maxall = maximum(s)
    rrange = max(abs(maxall), abs(minall))

    plt = volumeslices!(ax, x, y, z, s, colorrange = (-rrange, rrange), colormap = :seismic)

    Colorbar(fig[1, 2], plt.plots[2])

    sl_yz, sl_xz, sl_xy = lsgrid.sliders

    lift(sl_yz.value) do v
        plt[:update_yz][](findfirst(v .== x))
    end
    lift(sl_xz.value) do v
        plt[:update_xz][](findfirst(v .== y))
    end
    lift(sl_xy.value) do v
        plt[:update_xy][](findfirst(v .== z))
    end

    set_close_to!(sl_yz, (x[end] + x[1]) / 2)
    set_close_to!(sl_xz, (y[end] + y[1]) / 2)
    set_close_to!(sl_xy, (z[end] + z[1]) / 2)

    display(fig)

end

"""
    sdf_to_vtk(filename, s, x, y, z) 
	
Write the signed distance field `s` defined by `x`, `y` and `z` axes to a vtk structured point dataset.
"""
function sdf_to_vtk(filename, s, x, y, z) 

    vtkfile = vtk_grid(filename, x, y, z)
    vtkfile["s"] = s
    vtk_save(vtkfile)

end 


export sdfgen, display_sdf, sdf_to_vtk


end
